#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

typedef struct FileHeader
{
   char          fileId[4];
   uint16_t      fileVersion;
   uint8_t       unknown[10];
} FileHeader;

typedef struct Matrix4x4
{
   float       data[4][4];
} Matrix4x4;

typedef struct Vector3
{
   float      x;
   float      y;
   float      z;
} Vector3;

typedef struct Quaternion
{
   float      x;
   float      y;
   float      z;
   float      w;
} Quaternion;

typedef struct CompressedVector3
{
   int16_t      x;
   int16_t      y;
   int16_t      z;
} CompressedVector3;

typedef struct CompressedQuaternion
{
   int16_t      x;
   int16_t      y;
   int16_t      z;
   int16_t      w;
} CompressedQuaternion;

typedef struct VertexBuffer
{
   Vector3      position;
   uint8_t      normals[4];         // Commented out by chrrox
   uint16_t     uv[2];
   uint8_t      colour[4];         // Commented out by chrrox
   uint8_t      boneIndices[4];
   uint8_t      boneWeights[4];
} VertexBuffer;


typedef struct LOD
{
   uint16_t         vertexCount;
   VertexBuffer*    vertexBuffer/*[vertexCount]*/;
   uint32_t         faceCount;
   uint16_t*        faceIndices/*[faceCount]*/;
} LOD;

typedef struct Mesh
{
   uint8_t          textureNameLength;
   char*            textureName/*[textureNameLength]*/;
   uint16_t         flag;
   LOD              lodArray[3];
} Mesh;

typedef struct ModelFile
{
   FileHeader             fileHeader;

   uint8_t          boneCount;
   uint32_t*        boneHashArray/*[boneCount]*/;
   uint8_t          usedBoneCount;
   int8_t*          usedBoneArray/*[usedBoneCount]*/;

   uint32_t         totalVertices;
   uint32_t         totalFaces;
   uint16_t         meshCount;

   Mesh*            meshes/*[meshCount]*/;
} ModelFile;


// Reads the .PAC file and returns a ModelFile variable with all the information in it
ModelFile* readPacFile(char* pathToPacFile);

// Prints All the information stored in a ModelFile, pausing each time a texture name is found
void printPACFile(ModelFile* modelFile);

// Reads the .PAC file and returns a ModelFile variable with all the information in it
ModelFile* readPacFile(char* pathToPacFile);

// Prints the array "textureName" from each mesh from modelFile
void printTextureNames(ModelFile* modelFile);

// Frees the memory used to store the PAC file information
void freeModelFile(ModelFile* modelFile);

// Gets only the DISTINCT texture names from the .PAC file.
// Returns a array of strings, each position, is a texture name
// Send a variable in the int* textureCountReturn parameter to know how many textures were found
// Note: This function skips all the irrelevant information and only allocates memory to store the distinct texture names
char** getTexturesFromFile(char* pathToPacFile, int* textureCountReturn);


int main()
{
    ModelFile* modelFile = readPacFile("m0127_ship_0003.pac");

    printPACFile(modelFile);
    printTextureNames(modelFile);

    freeModelFile(modelFile);

    int textureCount = 0;

    char** textureNames = getTexturesFromFile("m0127_ship_0003.pac",&textureCount);

    int i = 0;
    for (i = 0; i < textureCount; i++)
    {
        printf("\n%d - %s", i + 1, textureNames[i]);
    }

    printf("\n\n");
    system("PAUSE");
    return 0;
}

ModelFile* readPacFile(char* pathToPacFile)
{
    printf("\nReading PAC file...");

    ModelFile* modelFile = (ModelFile*) malloc(1 * sizeof(ModelFile));

    FILE* fp = fopen(pathToPacFile,"rb");
    if (!fp)
    {
        printf("Failed to open file: %s", pathToPacFile);
        printf("\n\n");
        system("PAUSE");
        return modelFile;
    }

    fread(&modelFile->fileHeader,sizeof(FileHeader),1,fp);
    fread(&modelFile->boneCount,sizeof(uint8_t),1,fp);

    modelFile->boneHashArray = (uint32_t*) malloc(modelFile->boneCount * sizeof(uint32_t));
    fread(modelFile->boneHashArray,sizeof(uint32_t),modelFile->boneCount,fp);

    fread(&modelFile->usedBoneCount,sizeof(uint8_t),1,fp);
    modelFile->usedBoneArray = (int8_t*) malloc(modelFile->usedBoneCount * sizeof(int8_t));
    fread(modelFile->usedBoneArray,sizeof(int8_t),modelFile->usedBoneCount,fp);

    fread(&modelFile->totalVertices,sizeof(uint32_t),1,fp);
    fread(&modelFile->totalFaces,sizeof(uint32_t),1,fp);
    fread(&modelFile->meshCount,sizeof(uint16_t),1,fp);

    modelFile->meshes = (Mesh*) malloc(modelFile->meshCount * sizeof(Mesh));

    int currentMeshIDX = 0;
    for (currentMeshIDX = 0; currentMeshIDX < modelFile->meshCount; currentMeshIDX++)
    {
        Mesh* currentMesh = &modelFile->meshes[currentMeshIDX];

        fread(&currentMesh->textureNameLength,sizeof(uint8_t),1,fp);

        currentMesh->textureName = (char*) malloc (currentMesh->textureNameLength + 1);
        fread(currentMesh->textureName,currentMesh->textureNameLength,1,fp);

        currentMesh->textureName[currentMesh->textureNameLength] = '\0';

        fread(&currentMesh->flag,sizeof(uint16_t),1,fp);

        int currentLODIDX = 0;
        for (currentLODIDX = 0; currentLODIDX < 3; currentLODIDX++)
        {
            LOD* currentLOD = &currentMesh->lodArray[currentLODIDX];

            fread(&currentLOD->vertexCount,sizeof(uint16_t),1,fp);
            currentLOD->vertexBuffer = (VertexBuffer*) malloc(currentLOD->vertexCount * sizeof(VertexBuffer));
            fread(currentLOD->vertexBuffer,sizeof(VertexBuffer),currentLOD->vertexCount,fp);

            fread(&currentLOD->faceCount,sizeof(uint32_t),1,fp);
            currentLOD->faceIndices = (uint16_t*) malloc(currentLOD->faceCount * sizeof(uint16_t));
            fread(currentLOD->faceIndices,sizeof(uint16_t),currentLOD->faceCount,fp);
        }
    }
    printf("\nDone\n");
    fclose(fp);
    return modelFile;
}

// Prints All the information stored in a ModelFile, pausing each time a texture name is found
void printPACFile(ModelFile* modelFile)
{
    printf("\nFile Header:");
    printf("\nFile ID     : %s", modelFile->fileHeader.fileId);
    printf("\nFile Version: %d", modelFile->fileHeader.fileVersion);
    printf("\nUnknown     : ");
    int i = 0;
    for (i = 0; i < 10; i ++)
    {
        printf("%" PRIu8 " ", modelFile->fileHeader.unknown[i]);
    }

    printf("\n\n");
    printf("\nBone count: %" PRIu8,modelFile->boneCount);

    printf("\n\nBone Hash Array: {\n");
    for (i = 0; i < modelFile->boneCount; i ++)
    {
        printf("%" PRIu32, modelFile->boneHashArray[i]);
        if (i < modelFile->boneCount - 1)
        {
            printf(",");
        }
        else
        {
            printf("}");
        }
    }

    printf("\n");
    printf("\nUsed bone count: %" PRIu8,modelFile->usedBoneCount);

    printf("\n\nUsed Bone Array: {");
    for (i = 0; i < modelFile->usedBoneCount; i ++)
    {
        printf("%" PRIu8, modelFile->usedBoneArray[i]);
        if (i < modelFile->usedBoneCount - 1)
        {
            printf(",");
        }
        else
        {
            printf("}");
        }
    }
    printf("}");

    printf("\n");
    printf("\nTotal Vertices: %" PRIu32,modelFile->totalVertices);

    printf("\n");
    printf("\nTotal Faces: %" PRIu32,modelFile->totalFaces);

    printf("\n");
    printf("\nMesh count: %" PRIu16,modelFile->meshCount);

    int currentMeshIDX = 0;
    for (currentMeshIDX = 0; currentMeshIDX < modelFile->meshCount; currentMeshIDX++)
    {
        Mesh* currentMesh = &modelFile->meshes[currentMeshIDX];

        printf("\nMESH[%d]:", currentMeshIDX);

        printf("\nTexture Name Length: %" PRIu8,currentMesh->textureNameLength);
        printf("\nTexture Name       : %s",currentMesh->textureName);

        // PAUSES HERE
        printf("\n\n");
        system("PAUSE");

        printf("\n");
        printf("\nFlag: %" PRIu16, currentMesh->flag);

        printf("\n\n");
        int currentLODIDX = 0;
        for (currentLODIDX = 0; currentLODIDX < 3; currentLODIDX++)
        {
            LOD* currentLOD = &currentMesh->lodArray[currentLODIDX];

            printf("\nLOD[%d]:\n", currentLODIDX);
            printf("Vertex Count: %" PRIu16, currentLOD->vertexCount);
            printf("\n\nVertexBuffer: {");

            int currentVertexIDX;
            for (currentVertexIDX = 0; currentVertexIDX < currentLOD->vertexCount; currentVertexIDX++)
            {
                VertexBuffer* currentVertex = &currentLOD->vertexBuffer[currentVertexIDX];
                printf("\n\tVertex[%d]:", currentVertexIDX);
                printf("\n\tx: %f  y:%f  z: %f", currentVertex->position.x, currentVertex->position.y,currentVertex->position.z);
                printf("\n\tNormals: %"PRIu8" %"PRIu8" %"PRIu8" %"PRIu8, currentVertex->normals[0],currentVertex->normals[1],currentVertex->normals[2],currentVertex->normals[3]);
                printf("\n\tUV: %"PRIu8" %"PRIu8, currentVertex->uv[0],currentVertex->uv[1]);
                printf("\n\tColour: %"PRIu8" %"PRIu8" %"PRIu8" %"PRIu8, currentVertex->colour[0],currentVertex->colour[1],currentVertex->colour[2],currentVertex->colour[3]);
                printf("\n\tBone Indices: %"PRIu8" %"PRIu8" %"PRIu8" %"PRIu8, currentVertex->boneIndices[0],currentVertex->boneIndices[1],currentVertex->boneIndices[2],currentVertex->boneIndices[3]);
                printf("\n\tBone Weights: %"PRIu8" %"PRIu8" %"PRIu8" %"PRIu8, currentVertex->boneWeights[0],currentVertex->boneWeights[1],currentVertex->boneWeights[2],currentVertex->boneWeights[3]);
                printf("\n");
            }
            printf("}\n\n");
            printf("Faces Count: %" PRIu32, currentLOD->faceCount);

             printf("\nFaces Indices: {");
             int currentFaceIDX = 0;
             for (currentFaceIDX = 0; currentFaceIDX < currentLOD->faceCount; currentFaceIDX++)
             {
                printf("%" PRIu16,currentLOD->faceIndices[currentFaceIDX]);
                if (currentFaceIDX < currentLOD->faceCount - 1)
                {
                    printf(",");
                }
                else
                {
                    printf("}");
                }
             }
             printf("\n----------------------------------------\n");
        }
    }
}

// Prints the array "textureName" from each mesh from modelFile
void printTextureNames(ModelFile* modelFile)
{
    int currentMeshIDX = 0;
    for (currentMeshIDX = 0; currentMeshIDX < modelFile->meshCount; currentMeshIDX++)
    {
        Mesh* currentMesh = &modelFile->meshes[currentMeshIDX];
        printf("\n%d - %s",currentMeshIDX + 1, currentMesh->textureName);
    }
}

// Frees the memory used to store the PAC file information
void freeModelFile(ModelFile* modelFile)
{
    free(modelFile->boneHashArray);
    free(modelFile->usedBoneArray);

    free(modelFile->meshes);

    int currentMeshIDX = 0;
    for (currentMeshIDX = 0; currentMeshIDX < modelFile->meshCount; currentMeshIDX++)
    {
        Mesh* currentMesh = &modelFile->meshes[currentMeshIDX];
        free(currentMesh->textureName);

        int currentLODIDX = 0;
        for (currentLODIDX = 0; currentLODIDX < 3; currentLODIDX++)
        {
            LOD* currentLOD = &currentMesh->lodArray[currentLODIDX];

            //free(currentLOD->vertexBuffer); // Not working, causes crash

            free(currentLOD->faceIndices);
        }
    }
}

// Gets only the DISTINCT texture names from the .PAC file.
// Returns a array of strings, each position, is a texture name
// Send a variable in the int* textureCountReturn parameter to know how many textures were found
// Note: This function skips all the irrelevant information and only allocates memory to store the distinct texture names
char** getTexturesFromFile(char* pathToPacFile, int* textureCountReturn)
{
    (*textureCountReturn) = 0;
    printf("\nGetting Textures from \"%s\"...", pathToPacFile);

    FILE* fp = fopen(pathToPacFile,"rb");
    if (!fp)
    {
        printf("Failed to open file: %s", pathToPacFile);
        printf("\n\n");
        system("PAUSE");
        return NULL;
    }


    fseek(fp,sizeof(FileHeader),SEEK_CUR); // Skip File Header

    uint8_t boneCount = 0;

    fread(&boneCount,sizeof(uint8_t),1,fp);
    fseek(fp,boneCount * sizeof(uint32_t),SEEK_CUR); // Skip boneHashArray

    uint8_t usedBoneCount = 0;
    fread(&usedBoneCount,sizeof(uint8_t),1,fp);

    int bytesToSkip = 0;
    bytesToSkip += usedBoneCount * sizeof(int8_t); // Skip usedBoneArray
    bytesToSkip += sizeof(uint32_t); // Skip totalVertices
    bytesToSkip += sizeof(uint32_t); // Skip totalFaces

    fseek(fp,bytesToSkip,SEEK_CUR); // Skips those bytes

    uint16_t meshCount = 0;
    fread(&meshCount,sizeof(uint16_t),1,fp);

    char** textureNames = (char**) malloc(meshCount * sizeof(char*));
    int i = 0;
    for (i = 0; i < meshCount; i++)
    {
        textureNames[i] = NULL;
    }
    int textureCount = 0;

    int currentMeshIDX = 0;
    for (currentMeshIDX = 0; currentMeshIDX < meshCount; currentMeshIDX++)
    {
        uint8_t textureNameLength = 0;
        fread(&textureNameLength,sizeof(uint8_t),1,fp);

        char* textureName = (char*) malloc(textureNameLength + 1);
        fread(textureName,textureNameLength,1,fp);

        textureName[textureNameLength] = '\0';
        textureCount++;

        for (i = 0; i < textureCount; i++)
        {
            if (textureNames[i] == NULL)
            {
                textureNames[i] = textureName;
                break;
            }
            else
            {
                if (strcmp(textureNames[i],textureName) == 0)
                {
                    textureCount--;
                    break;
                }
            }
        }

        uint16_t flag = 0;
        fread(&flag,sizeof(uint16_t),1,fp);

        int currentLODIDX = 0;
        for (currentLODIDX = 0; currentLODIDX < 3; currentLODIDX++)
        {
            uint16_t vertexCount;
            fread(&vertexCount,sizeof(uint16_t),1,fp);

            fseek(fp,vertexCount * sizeof(VertexBuffer),SEEK_CUR); // Skips vertexBuffer

            uint32_t faceCount = 0;
            fread(&faceCount,sizeof(uint32_t),1,fp);

            fseek(fp,faceCount * sizeof(uint16_t),SEEK_CUR); // Skips faceIndices
        }
    }
    fclose(fp);
    (*textureCountReturn) = textureCount;

    printf("\nDone\n");

    return textureNames;
}
